using EPiServer.Security;
using EPiServer.ServiceLocation;
using Mediachase.Commerce.Catalog;
using Mediachase.Commerce.Catalog.Managers;
using Mediachase.Commerce.Catalog.Objects;
using Mediachase.Commerce.Core;
using Mediachase.Commerce.Core.Features;
using Mediachase.Commerce.Engine;
using Mediachase.Commerce.Extensions;
using Mediachase.Commerce.Inventory;
using Mediachase.Commerce.InventoryService;
using Mediachase.Commerce.Marketing;
using Mediachase.Commerce.Marketing.Dto;
using Mediachase.Commerce.Marketing.Managers;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.Orders.Managers;
using Mediachase.Commerce.Pricing;
using Mediachase.Commerce.Security;
using Mediachase.MetaDataPlus.Configurator;
using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Threading;

namespace Mediachase.Commerce.Website.Helpers
{
    /// <summary>
    /// Cart helper class used to simplify cart operations.
    /// The cart is automatically cached in the current Http Context.
    /// </summary>
    public class CartHelper
    {
        private static Injected<ICurrentMarket> CurrentMarketService { get; set; }
        private static Injected<IPriceService> PriceService { get; set; }
        private static Injected<IWarehouseRepository> WarehouseRepository { get; set; }
        private static Injected<IInventoryService> InventoryService { get; set; }
        private static Injected<IFeatureSwitch> FeatureSwitch { get; set; }

        /// <summary>
        /// Default name for the cart.
        /// </summary>
        public static string WishListName = Cart.WishListName;
        private string _entryDescriptionMetaFieldName = "Description";

        #region Data Members
        private Cart _Cart;
        #endregion

        #region Constructors
        /// <summary>
        /// Initializes a new instance of the <see cref="CartHelper"/> class.
        /// </summary>
        /// <param name="name">The name.</param>
        public CartHelper(string name)
        {
            Guid userId = PrincipalInfo.CurrentPrincipal.GetContactId();
            LoadCart(name, userId);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CartHelper"/> class.
        /// </summary>
        /// <param name="name">The name of cart</param>
        /// <param name="userId">The User Id.</param>
        /// <param name="market">The market.</param>
        public CartHelper(string name, Guid userId, IMarket market)
        {
            LoadCart(name, userId, market.MarketId);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CartHelper"/> class.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="userId">The user id.</param>
        public CartHelper(string name, Guid userId)
        {
            LoadCart(name, userId);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CartHelper"/> class.
        /// </summary>
        /// <param name="cart">The cart.</param>
        public CartHelper(Cart cart)
        {
            if (cart == null)
            {
                throw new ArgumentNullException("cart");
            }
            this._Cart = cart;
        }
        #endregion

        #region Public properties
        /// <summary>
        /// Gets the cart.
        /// </summary>
        /// <value>The cart.</value>
        public virtual Cart Cart
        {
            get
            {
                return this._Cart;
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is empty.
        /// </summary>
        /// <value><c>true</c> if this instance is empty; otherwise, <c>false</c>.</value>
        public virtual bool IsEmpty
        {
            get
            {
                foreach (OrderForm orderForm in this.Cart.OrderForms)
                {
                    if (orderForm.LineItems.Count > 0)
                    {
                        return false;
                    }
                }

                return true;
            }
        }

        /// <summary>
        /// Gets the line items.
        /// </summary>
        /// <value>The line items.</value>
        public virtual IEnumerable<LineItem> LineItems
        {
            get
            {
                foreach (OrderForm orderForm in this.Cart.OrderForms)
                {
                    foreach (LineItem lineItem in orderForm.LineItems)
                    {
                        yield return lineItem;
                    }
                }
            }
        }

        /// <summary>
        /// Gets the primary address.
        /// </summary>
        /// <value>The primary address.</value>
        public virtual OrderAddress PrimaryAddress
        {
            get
            {
                if (String.IsNullOrEmpty(OrderForm.BillingAddressId))
                    return null;

                foreach (OrderAddress address in Cart.OrderAddresses)
                {
                    if (address.Name.Equals(OrderForm.BillingAddressId, StringComparison.OrdinalIgnoreCase))
                        return address;
                }

                return null;
            }
        }

        /// <summary>
        /// Gets the order form.
        /// </summary>
        /// <value>The order form.</value>
        public virtual OrderForm OrderForm
        {
            get
            {
                return GetOrderForm();
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is address required.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if this instance is address required; otherwise, <c>false</c>.
        /// </value>
        public virtual bool IsAddressRequired
        {
            get
            {
                foreach (OrderForm orderForm in this.Cart.OrderForms)
                {
                    foreach (Shipment shipment in orderForm.Shipments)
                    {
                        if (String.IsNullOrEmpty(shipment.ShippingAddressId))
                        {
                            return true;
                        }
                    }
                }

                return false;
            }
        }

        public string EntryDescriptionMetaFieldName
        {
            get { return _entryDescriptionMetaFieldName; }
            set { _entryDescriptionMetaFieldName = value; }
        }
        #endregion

        #region Public methods
        /// <summary>
        /// Finds the address by id.
        /// </summary>
        /// <param name="addressId">The address id.</param>
        /// <returns></returns>
        public virtual OrderAddress FindAddressById(string addressId)
        {
            if (String.IsNullOrEmpty(addressId))
                return null;

            foreach (OrderAddress address in Cart.OrderAddresses)
            {
                if (address.OrderGroupAddressId == Int32.Parse(addressId))
                    return address;
            }

            return null;
        }

        /// <summary>
        /// Finds the name of the address by.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns></returns>
        public virtual OrderAddress FindAddressByName(string name)
        {
            if (name == null)
                return null;

            foreach (OrderAddress address in Cart.OrderAddresses)
            {
                if (address.Name.ToUpper() == name.ToUpper())
                    return address;
            }

            return null;
        }

        /// <summary>
        /// Gets the total number of items in the basket.
        /// </summary>
        /// <returns>the total number of items in the basket.</returns>
        public virtual decimal GetTotalItemCount()
        {
            decimal totalItemCount = 0;
            foreach (OrderForm orderForm in this.Cart.OrderForms)
            {
                foreach (LineItem lineItem in orderForm.LineItems)
                {
                    totalItemCount += lineItem.Quantity;
                }
            }
            return totalItemCount;
        }

        /// <summary>
        /// Gets the default OrderForm.
        /// </summary>
        /// <returns>the default OrderForm</returns>
        public virtual OrderForm GetOrderForm()
        {
            return this.GetOrderForm(Cart.Name);
        }

        /// <summary>
        /// Gets the named OrderForm.
        /// </summary>
        /// <param name="orderFormName">The name of the OrderForm object to retrieve.</param>
        /// <returns>The named OrderForm.</returns>
        public virtual OrderForm GetOrderForm(string orderFormName)
        {
            if (String.IsNullOrEmpty(orderFormName))
            {
                orderFormName = Cart.DefaultName;
            }

            OrderForm orderForm = this.Cart.OrderForms[orderFormName];
            if (orderForm == null)
            {
                orderForm = new OrderForm();
                orderForm.Name = orderFormName;
                this.Cart.OrderForms.Add(orderForm);
            }
            return orderForm;
        }

        /// <summary>
        /// Deletes the current basket instance from the database.
        /// </summary>
        public virtual void Delete()
        {
            // Remove any reservations
            // Load existing usage Dto for the current order
            PromotionUsageDto usageDto = PromotionManager.GetPromotionUsageDto(0, Guid.Empty, this.Cart.OrderGroupId);

            // Clear all old items first
            if (usageDto.PromotionUsage.Count > 0)
            {
                foreach (PromotionUsageDto.PromotionUsageRow row in usageDto.PromotionUsage)
                {
                    if (row.OrderGroupId == this.Cart.OrderGroupId)
                    {
                        row.Delete();
                    }
                }
            }

            // Save the promotion usage
            PromotionManager.SavePromotionUsage(usageDto);

            string name = this.Cart.Name;

            // Delete the cart
            this.Cart.Delete();
        }

        /// <summary>
        /// Adds the entry. Line item's qty will be increased by 1.
        /// </summary>
        /// <param name="entry">The entry.</param>
        /// <returns>The line item.</returns>
        public virtual LineItem AddEntry(Entry entry)
        {
            return AddEntry(entry, 1, false);
        }

        /// <summary>
        /// Adds the entry. Line item's qty will be increased by 1.
        /// </summary>
        /// <param name="entry">The entry.</param>
        /// <param name="helpersToRemove"></param>
        /// <returns>The line item.</returns>
        public virtual LineItem AddEntry(Entry entry, params CartHelper[] helpersToRemove)
        {
            return AddEntry(entry, 1, false, helpersToRemove);
        }

        /// <summary>
        /// Adds the entry.
        /// </summary>
        /// <param name="entry">The entry.</param>
        /// <param name="fixedQuantity">If true, lineitem's qty will be set to <paramref name="fixedQuantity"/> value. Otherwise, <paramref name="fixedQuantity"/> will be added to the current line item's qty value.</param>
        /// <returns>The line item.</returns>
        public virtual LineItem AddEntry(Entry entry, bool fixedQuantity)
        {
            return AddEntry(entry, 1, fixedQuantity);
        }

        /// <summary>
        /// Adds the entry.
        /// </summary>
        /// <param name="entry">The entry.</param>
        /// <param name="fixedQuantity">If true, lineitem's qty will be set to <paramref name="fixedQuantity"/> value. Otherwise, <paramref name="fixedQuantity"/> will be added to the current line item's qty value.</param>
        /// <param name="helpersToRemove"></param>
        /// <returns>The line item.</returns>
        public virtual LineItem AddEntry(Entry entry, bool fixedQuantity, params CartHelper[] helpersToRemove)
        {
            return AddEntry(entry, 1, fixedQuantity, helpersToRemove);
        }

        /// <summary>
        /// Adds the entry with default warehouse code
        /// </summary>
        /// <param name="entry">The entry.</param>
        /// <param name="quantity">The quantity.</param>
        /// <param name="fixedQuantity">If true, lineitem's qty will be set to <paramref name="quantity"/> value. Otherwise, <paramref name="quantity"/> will be added to the current line item's qty value.</param>
        /// <param name="helpersToRemove">CartHelper(s) from which the item needs to be removed simultaneously with adding it to the current CartHelper.</param>
        /// <returns>The line item.</returns>
        public virtual LineItem AddEntry(Entry entry, decimal quantity, bool fixedQuantity, params CartHelper[] helpersToRemove)
        {
            return AddEntry(entry, quantity, fixedQuantity, String.Empty, helpersToRemove);
        }

        /// <summary>
        /// Adds the entry.
        /// </summary>
        /// <param name="entry">The entry.</param>
        /// <param name="quantity">The quantity.</param>
        /// <param name="fixedQuantity">If true, lineitem's qty will be set to <paramref name="quantity"/> value. Otherwise, <paramref name="quantity"/> will be added to the current line item's qty value.</param>
        /// <param name="warehouseCode">The warehouse code</param>
        /// <param name="helpersToRemove">CartHelper(s) from which the item needs to be removed simultaneously with adding it to the current CartHelper.</param>
        /// <returns>The line item.</returns>
        public virtual LineItem AddEntry(Entry entry, decimal quantity, bool fixedQuantity, string warehouseCode, params CartHelper[] helpersToRemove)
        {
            if (entry.EntryType == EntryType.Bundle)
            {
                // Add items of bundle to cart.
                var bundle = LoadCatalogEntryWithChildren(entry.CatalogEntryId);

                return AddEntry(bundle.Entries.Entry, bundle.Entries.Entry.Select(e => e.RelationInfo.Quantity * quantity).ToArray(), fixedQuantity, warehouseCode, helpersToRemove);
            }
            else
            {
                // Add line items
                return AddEntry(new Entry[] { entry }, new decimal[] { quantity }, fixedQuantity, warehouseCode, helpersToRemove);
            }
        }

        private LineItem AddEntry(Entry[] entries, decimal[] quantity, bool fixedQuantity, string warehouseCode, params CartHelper[] helpersToRemove)
        {
            Guid guid = PrincipalInfo.CurrentPrincipal.GetContactId();// this should persist in the cookie
            LineItem retLineItem = null;
            this.Cart.BillingCurrency = SiteContext.Current.Currency;

            OrderForm orderForm = null;
            if (this.Cart.OrderForms.Count == 0) // create a new one
            {
                orderForm = new OrderForm();
                orderForm.Name = this.Cart.Name;
                this.Cart.OrderForms.Add(orderForm);
            }
            else // use first one
            {
                orderForm = this.Cart.OrderForms[0];
            }

            // Add line items
            for (var i = 0; i < entries.Count(); i++)
            {
                var entry = entries[i];
                LineItem lineItem = CreateLineItem(entry, quantity[i], warehouseCode);

                // check if items already exist
                bool found = false;
                foreach (LineItem item in orderForm.LineItems)
                {
                    if (item.Code == lineItem.Code && !item.IsGift)
                    {
                        item.Quantity = fixedQuantity ? lineItem.Quantity : item.Quantity + lineItem.Quantity;
                        item.ExtendedPrice = lineItem.ListPrice;
#pragma warning disable 618
                        item.WarehouseCode = warehouseCode;
#pragma warning restore 618
                        item.InventoryStatus = lineItem.InventoryStatus;
                        retLineItem = item;
                        found = true;
                        break;
                    }
                }

                if (!found)
                {
                    orderForm.LineItems.Add(lineItem);
                    retLineItem = lineItem;
                }
            }

            // remove entry from other helpers if needed
            if (helpersToRemove != null && helpersToRemove.Length > 0)
            {
                foreach (CartHelper ch in helpersToRemove)
                {
                    var changed = false;
                    // if entry is in the helper, remove it
                    var lineItems = ch.LineItems.Where(item => entries.Select(e => e.ID).Contains(item.Code));
                    foreach (LineItem lineItem in lineItems)
                    {
                        if (lineItem != null)
                        {
                            lineItem.Delete();

                            // If helper is empty, remove it from the database
                            if (ch.IsEmpty)
                                ch.Delete();

                            changed = true;
                        }
                    }

                    // save changes
                    if (changed)
                    {
                        ch.Cart.AcceptChanges();
                    }
                }
            }

            // Save cart
            this.Cart.AcceptChanges();
            if (FeatureSwitch.Service.IsVNextWorkflowEnabled())
            {
                for (var i = 0; i < entries.Count(); i++)
                {
                    for (var j = 0; j < orderForm.LineItems.Count; j++)
                    {
                        var lineItem = orderForm.LineItems[j];
                        if (lineItem.Code != entries[i].ID)
                        {
                            continue;
                        }

                        decimal quantityToAddForShipment;
                        if (fixedQuantity)
                        {
                            var currentQuantityForShipment = GetQuantityForLineItemOnFirstShipment(lineItem.LineItemId);
                            quantityToAddForShipment = quantity[i] - currentQuantityForShipment;
                        }
                        else
                        {
                            quantityToAddForShipment = quantity[i];
                        }

                        AddItemToShipment(j, quantityToAddForShipment, warehouseCode);
                    }
                }
            }

            return retLineItem;
        }

        /// <summary>
        /// Create the lineitem
        /// </summary>
        /// <param name="entry">The entry</param>
        /// <param name="quantity">The quantity</param>
        /// <param name="warehouseCode">The warehouse code</param>
        /// <returns></returns>
        private LineItem CreateLineItem(Entry entry, decimal quantity, string warehouseCode)
        {
            LineItem lineItem = new LineItem();

            // If entry has a parent, add parents name
            if (entry.ParentEntry != null)
            {
                lineItem.DisplayName = String.Format("{0}: {1}", entry.ParentEntry.Name, entry.Name);
                lineItem.ParentCatalogEntryId = entry.ParentEntry.ID;
            }
            else
            {
                lineItem.DisplayName = entry.Name;
                lineItem.ParentCatalogEntryId = String.Empty;
            }

            MarketId currentMarketId = CurrentMarketService.Service.GetCurrentMarket().MarketId;
            IPriceValue listPrice = PriceService.Service.GetDefaultPrice(currentMarketId, FrameworkContext.Current.CurrentDateTime, new CatalogKey(entry), SiteContext.Current.Currency);

            lineItem.Code = entry.ID;
            lineItem.MaxQuantity = entry.ItemAttributes.MaxQuantity;
            lineItem.MinQuantity = entry.ItemAttributes.MinQuantity;
            lineItem.Quantity = quantity;

            if (String.IsNullOrEmpty(warehouseCode))
            {
                lineItem.InventoryStatus = (int)entry.InventoryStatus;
            }
            else
            {
#pragma warning disable 618
                lineItem.WarehouseCode = warehouseCode;
#pragma warning restore 618
                lineItem.InventoryStatus = (int)GetInventoryStatus(entry, warehouseCode);
            }

            if (listPrice != null)
            {
                lineItem.ListPrice = listPrice.UnitPrice.Amount;
                lineItem.PlacedPrice = listPrice.UnitPrice.Amount;
                lineItem.ExtendedPrice = listPrice.UnitPrice.Amount;
            }
            else
            {
                lineItem.ListPrice = lineItem.PlacedPrice;
            }

            //Populate LineItem description field from the Entry object, if it exists in entry metaclass
            //Entry description metafield is also configurable so it could map another field to the description field . This fixes issue #718
            if (!string.IsNullOrEmpty(EntryDescriptionMetaFieldName))
            {
                MetaClass mc = MetaClass.Load(CatalogContext.MetaDataContext, entry.MetaClassId);
                foreach (MetaField mf in mc.MetaFields)
                {
                    if (mf.Name.Equals(EntryDescriptionMetaFieldName, StringComparison.InvariantCultureIgnoreCase) &&
                        entry.ItemAttributes[EntryDescriptionMetaFieldName] != null)
                    {
                        //description mapped field found - add to LineItem.Description field
                        lineItem.Description = entry.ItemAttributes[EntryDescriptionMetaFieldName].ToString();
                    }
                }
            }

            return lineItem;
        }

        private InventoryTrackingStatus GetInventoryStatus(Entry entry, string warehouseCode)
        {
            var inventoryStatus = entry.InventoryStatus;
            var inventory = InventoryService.Service.Get(entry.ID, warehouseCode);
            if (inventory != null)
            {
                inventoryStatus = (InventoryTrackingStatus)Convert.ToInt32(inventory.IsTracked);
            }
            
            return inventoryStatus;
        }

        private Entry LoadCatalogEntryWithChildren(int catalogEntryId)
        {
            CatalogEntryResponseGroup responseGroup = new CatalogEntryResponseGroup(CatalogEntryResponseGroup.ResponseGroup.CatalogEntryInfo | CatalogEntryResponseGroup.ResponseGroup.Children);

            return CatalogContext.Current.GetCatalogEntry(catalogEntryId, responseGroup);
        }

        /// <summary>
        /// Runs the workflow and generates the error message for all the warnings.
        /// </summary>
        /// <param name="name">The name.</param>
        public virtual void RunWorkflow(string name)
        {
            RunWorkflow(name, null);
        }

        /// <summary>
        /// Runs the workflow and generates the error message for all the warnings.
        /// </summary>
        /// <param name="name">The workflow name.</param>
        /// <param name="param">The workflow parameters.</param>
        public virtual void RunWorkflow(string name, Dictionary<string, object> param)
        {
            WorkflowResults results = OrderGroupWorkflowManager.RunWorkflow(this.Cart, name, true, param);

            object warnings = results.OutputParameters["Warnings"];
            if (warnings != null)
            {
                StringDictionary warningsList = warnings as StringDictionary;
                if (warningsList != null)
                {
                    foreach (string warning in warningsList.Values)
                    {
                        ErrorManager.GenerateError(warning);
                    }
                }
            }
        }

        /// <summary>
        /// Gets all of the cart promotions regardless of promotion type.
        /// </summary>
        /// <returns></returns>
        public IEnumerable<string> GetCartPromotions()
        {
            return GetCartPromotions(null);
        }

        /// <summary>
        /// Gets the cart promotions based on promotion type or All.
        /// </summary>
        /// <param name="promoKey">The promotion group key.</param>
        /// <returns></returns>
        public IEnumerable<string> GetCartPromotions(PromotionGroup.PromotionGroupKey? promoKey)
        {
            IEnumerable<Discount> discounts = null;
            OrderForm[] allOrderForms = new OrderForm[] { this.OrderForm };
            IEnumerable<Shipment> allShipments;
            IEnumerable<LineItem> allLineItems;

            switch (promoKey)
            {
                case PromotionGroup.PromotionGroupKey.Order:
                    discounts = allOrderForms.SelectMany(x => x.Discounts.Cast<Discount>());
                    break;
                case PromotionGroup.PromotionGroupKey.Shipping:
                    allShipments = allOrderForms.SelectMany(x => x.Shipments.ToArray());
                    discounts = allShipments.SelectMany(x => x.Discounts.Cast<Discount>());
                    break;
                case PromotionGroup.PromotionGroupKey.Entry:
                    allLineItems = allOrderForms.SelectMany(x => x.LineItems.ToArray());
                    discounts = allLineItems.SelectMany(x => x.Discounts.Cast<Discount>());
                    break;
                default:
                    // Get all discounts
                    allShipments = allOrderForms.SelectMany(x => x.Shipments.ToArray());
                    allLineItems = allOrderForms.SelectMany(x => x.LineItems.ToArray());
                    discounts = allOrderForms.SelectMany(x => x.Discounts.Cast<Discount>());
                    discounts = discounts.Concat(allShipments.SelectMany(x => x.Discounts.Cast<Discount>()));
                    discounts = discounts.Concat(allLineItems.SelectMany(x => x.Discounts.Cast<Discount>()));
                    break;
            }

            //Grouping discounts by promotion type
            var groupingDiscounts = discounts.GroupBy(x => x.DiscountId);

            PromotionDto dto = PromotionManager.GetPromotionDto();
            foreach (var groupDiscount in groupingDiscounts)
            {
                var promotionId = groupDiscount.Key;
                PromotionDto.PromotionRow promotion = dto.Promotion.SingleOrDefault(x => x.PromotionId == promotionId);
                if (promotion != null)
                {
                    var retVal = GetPromotionDisplayName(promotion, Thread.CurrentThread.CurrentCulture.Name);
                    if (string.IsNullOrEmpty(retVal))
                    {
                        retVal = promotion.Name;
                    }
                    yield return retVal;
                }
            }
        }

        private static string GetPromotionDisplayName(PromotionDto.PromotionRow row, string languageCode)
        {
            PromotionDto.PromotionLanguageRow[] langRows = row.GetPromotionLanguageRows();
            if (langRows != null && langRows.Length > 0)
            {
                foreach (PromotionDto.PromotionLanguageRow lang in langRows)
                {
                    if (lang.LanguageCode.Equals(languageCode, StringComparison.OrdinalIgnoreCase))
                    {
                        return lang.DisplayName;
                    }
                }
            }

            return row.Name;
        }

        /// <summary>
        /// Resets this instance. Will clean up line items, remove payments and delete addresses.
        /// The cart needs to be saved in order for changes to be persisted.
        /// </summary>
        public virtual void Reset()
        {
            // Reset shopping cart
            foreach (LineItem lineItem in this.LineItems)
            {
#pragma warning disable 618
                lineItem.ShippingAddressId = String.Empty;
                lineItem.ShippingMethodId = Guid.Empty;
                lineItem.ShippingMethodName = null;
#pragma warning restore 618
            }

            foreach (OrderForm orderForm in Cart.OrderForms)
            {
                foreach (Payment orderPayment in orderForm.Payments)
                {
                    orderPayment.Delete();
                }
                if (FeatureSwitch.Service.IsVNextWorkflowEnabled())
                {
                    MergeShipments(orderForm);
                    ClearShippingInfo(orderForm);
                }
                else
                {
                    foreach (Shipment shipment in orderForm.Shipments)
                    {
                        shipment.Delete();
                    }
                }
            }

            foreach (OrderAddress orderAddress in Cart.OrderAddresses)
            {
                orderAddress.Delete();
            }

            Cart.AddressId = String.Empty;
        }

        /// <summary>
        /// Clears the shipping information.
        /// </summary>
        /// <param name="orderForm">The order form.</param>
        public virtual void ClearShippingInfo(OrderForm orderForm)
        {
            foreach (var shipment in orderForm.Shipments.ToList())
            {
                shipment.ShippingAddressId = string.Empty;
                shipment.ShippingMethodId = Guid.Empty;
                shipment.ShippingMethodName = string.Empty;
            }
        }

        private decimal GetQuantityForLineItemOnFirstShipment(int lineItemId)
        {
            var shipment = OrderForm.Shipments.FirstOrDefault();
            if (shipment == null)
            {
                shipment = OrderForm.Shipments.AddNew();
            }

            return Shipment.GetLineItemQuantity(shipment, lineItemId);
        }

        [Obsolete("This method is no longer called from any EPiServer API. Use 'AddItemToShipment' with warehouseCode parameter instead. Will remain at least until October 2017.")]
        protected void AddItemToFirstShipment(int lineItemIndex, decimal quantity)
        {
            AddItemToShipment(lineItemIndex, quantity, string.Empty);
        }

        protected void AddItemToShipment(int lineItemIndex, decimal quantity, string warehouseCode)
        {
            var shipment = OrderForm.Shipments.FirstOrDefault();
            if (shipment == null || shipment.WarehouseCode != warehouseCode)
            {
                shipment = OrderForm.Shipments.AddNew();
                shipment.WarehouseCode = warehouseCode;
            }

            shipment.AddLineItemIndex(lineItemIndex, quantity);
            Cart.AcceptChanges();
        }

        [Obsolete("This method is no longer called from any EPiServer API. Use 'AddItemToFirstShipment' instead. Will remain at least until July 2016.")]
        protected void AddLastItemToFirstShipment()
        {
            AddItemToFirstShipment(LineItems.Count() - 1, LineItems.Last().Quantity);
        }

        /// <summary>
        /// Merge all shipments of order forms into one for each form.
        /// </summary>
        public void MergeShipments()
        {
            if (Cart.OrderForms != null)
            {
                foreach (OrderForm form in Cart.OrderForms)
                {
                    MergeShipments(form);
                }
            }
        }

        /// <summary>
        /// Merge all shipments of an order form.
        /// </summary>
        /// <param name="form">The order form</param>
        public void MergeShipments(OrderForm form)
        {
            var shipments = form.Shipments;
            if (shipments.Count <= 1)
            {
                return;
            }

            var shipmentsToRemove = new List<Shipment>();
            for (int i = 0; i < shipments.Count - 1; i++)
            {
                if (!shipments[i].LineItemIndexes.Any())
                {
                    shipmentsToRemove.Add(shipments[i]);
                    continue;
                }

                for (int j = i + 1; j < shipments.Count; j++)
                {
                    if (shipments[i].WarehouseCode == shipments[j].WarehouseCode)
                    {
                        MergeShipment(form, shipments[j], shipments[i]);
                        shipmentsToRemove.Add(shipments[j]);
                    }
                }
            }

            foreach (var shipment in shipmentsToRemove)
            {
                shipment.Delete();
            }
        }

        private void MergeShipment(OrderForm form, Shipment source, Shipment destination)
        {
            foreach (var itemIndex in source.LineItemIndexes)
            {
                var lineItem = form.LineItems[int.Parse(itemIndex)];
                if (lineItem != null)
                {
                    // Get quantity with line item id.
                    destination.AddLineItemIndex(int.Parse(itemIndex), Shipment.GetLineItemQuantity(source, lineItem.LineItemId));
                    source.RemoveLineItemIndex(itemIndex);
                }
            }
        }

        #endregion

        #region Protected Methods
        /// <summary>
        /// Gets the cache key which is used to store current cart in the http context, by a specific market.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="userId">The user id.</param>
        /// <param name="marketId">The market id.</param>
        /// <returns></returns>
        protected virtual string GetCacheKey(string name, Guid userId, MarketId marketId)
        {
            if (name == null)
            {
                throw new ArgumentNullException(nameof(name));
            }

            if (userId == Guid.Empty)
            {
                throw new ArgumentException("Invalid user id, can't be empty", nameof(userId));
            }

            if (marketId == null)
            {
                throw new ArgumentNullException(nameof(marketId));
            }

            // Cache cart in the current context
            string cartKey = OrderCache.CreateCacheKey("Cart-http", name, userId.ToString(), marketId.ToString());

            return cartKey;
        }


        /// <summary>
        /// Loads the cart. The cart is loaded from current http context if one is present.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="userId">The user id.</param>
        /// <remarks>Uses the market returned from ICurrentMarket for a market id.</remarks>
        protected virtual void LoadCart(string name, Guid userId)
        {
            ICurrentMarket currentMarketService = ServiceLocator.Current.GetInstance<ICurrentMarket>();
            MarketId currentMarketId = currentMarketService.GetCurrentMarket().MarketId;

            LoadCart(name, userId, currentMarketId);
        }


        /// <summary>
        /// Loads the cart. The cart is loaded from current http context if one is present.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="userId">The user id.</param>
        /// <param name="marketId">The market id.</param>
        protected virtual void LoadCart(string name, Guid userId, MarketId marketId)
        {
            var cart = OrderContext.Current.GetCart(name, userId, marketId);

            if (String.IsNullOrEmpty(cart.CustomerName) || (cart.CustomerName.Equals(SecurityContext.Anonymous, StringComparison.OrdinalIgnoreCase) && PrincipalInfo.CurrentPrincipal.Identity.IsAuthenticated))
            {
                cart.CustomerName = PrincipalInfo.CurrentPrincipal.Identity.Name;
            }

            if (FeatureSwitch.Service.IsVNextWorkflowEnabled())
            {
                var defaultOrderForm = cart.OrderForms.FirstOrDefault();
                if (defaultOrderForm == null)
                {
                    defaultOrderForm = cart.OrderForms.AddNew();
                    defaultOrderForm.Name = cart.Name;
                }
            }

            this._Cart = cart;
        }
        #endregion
    }
}
